/*
 * gencall.h
 *				    Gianni Mariani 3-Dec-1993
 *
 *  Generalized calling function.  This allows the calling
 *  of any function adhering to the MIPS calling conventions.
 *  It manipulates the CPU registers to adhere to the calling
 *  standards.  This allows high level language programs to call
 *  any function at run time by assembling the parameters in the
 *  stack area given to a preparation function.
 *
 *  CallFromGeneralArgs is an assembler routine that generates
 *  the arg stack, calls a prep routine and then calls the function
 *  in question.  The prep routine may be a simple memcpy() of the
 *  args being passed into the arguments area allocated by
 *  CallFromGeneralArgs.
 *
 *  gcif_Prefunc takes a call description and prepares arguments
 *  in a pointer passed to it.  It acts as a prepare function
 *  to CallFromGeneralArgs.
 *
 *  gcif_call takes a function address and a call description and
 *  calls that function with that description.
 *
 */

/* flags values								*/
#define RETVOID     0
#define RETINT      1
#define RETFLOAT    2
#define RETDOUBLE   4
#define RET_FP_MASK ( RETFLOAT | RETDOUBLE )
#define RETMASK	    0xffff

#define PASSSHIFT   16

#define XPASS_FLOAT  RETFLOAT
#define XPASS_DOUBLE RETDOUBLE
#define XPASS_SHIFT  4

#define	PASS_D	    XPASS_DOUBLE
#define	PASS_F	    XPASS_FLOAT
#define	PASS_D_D    XPASS_DOUBLE + XPASS_DOUBLE * XPASS_SHIFT
#define	PASS_D_F    XPASS_DOUBLE + XPASS_FLOAT  * XPASS_SHIFT
#define	PASS_F_F    XPASS_FLOAT  + XPASS_FLOAT  * XPASS_SHIFT
#define	PASS_F_D    XPASS_FLOAT  + XPASS_DOUBLE * XPASS_SHIFT

#define	FP_PASS_D    ( ( PASS_D )   << PASSSHIFT )
#define	FP_PASS_F    ( ( PASS_F )   << PASSSHIFT )
#define	FP_PASS_D_D  ( ( PASS_D_D ) << PASSSHIFT )
#define	FP_PASS_D_F  ( ( PASS_D_F ) << PASSSHIFT )
#define	FP_PASS_F_F  ( ( PASS_F_F ) << PASSSHIFT )
#define	FP_PASS_F_D  ( ( PASS_F_D ) << PASSSHIFT )

#define FP_PASS( A, B )	((A) ? ( ((A)+(B) * XPASS_SHIFT)<<PASSSHIFT ) : 0)

/* ==================================================================== */
/* This is only needed if you are using gcif_Prefunc as a prepare func	*/

#ifndef GCIF_ASHEADER

/* can be used to define gcif_flags					*/
/* e.g. GCIF_MKFLAGS( gcif_double, FP_PASS( gcif_double, gcif_void ) )	*/
/* defines a function returning a double being passed a double in the	*/
/* first arg.								*/
#define GCIF_MKFLAGS( RETTYPE, PASSFLAGS )  ((RETTYPE)|(PASSFLAGS))


typedef void	* GCIF_FUNC_T;


/* ======== CallFromGeneralArgs ======================================= */
/* PURPOSE:
 *	calls func with any arguments.
 *
 * RETURNS:
 * 	Whatever func returns except for structs it returns void.
 *	Struct return values are actually pointers passed as the first arg
 *	to func.
 */

int CallFromGeneralArgs(
    GCIF_FUNC_T	    Prefunc,	    /* Address of call args prep func	*/
    void          * args,	    /* Parameter for Prefunc		*/
    unsigned        bytes,	    /* Parameter for Prefunc		*/
    unsigned        flags,	    /* passing and return value flags	*/
    void          * retval,	    /* pointer to return value - opt.	*/
    GCIF_FUNC_T     func	    /* address of function to call	*/
);


typedef struct { char v[4]; } GCIF_UAS;

/* ======== GCIF_ETYPE ================================================ */
/* PURPOSE:
 *	
 */

typedef enum GCIF_ETYPE {

    /* gcif_void is for when you don't return ar pass anything		*/
    gcif_void = RETVOID,    /* void return value			*/

    /* int types include pointers or any 32 bit interger value		*/
    gcif_int = RETINT,	    /* Integer, pointer etc			*/
    gcif_float = RETFLOAT,  /* float arg				*/
    gcif_double = RETDOUBLE,/* double arg				*/

    /* gcif_struct types are obvious					*/
    gcif_struct = 16,	/* struct arg - basically an int arg		*/

    /* the integer types of sizes less than 32 bits need conversion	*/
    /* to a 32 bit type when being passed.  This is why they need	*/
    /* special treatment.						*/
    gcif_uchar = 16*2,
    gcif_schar = 16*3,
    gcif_ushort = 16*4,
    gcif_sshort = 16*5,
    gcif_char = 16*6

} GCIF_ETYPE;

typedef unsigned char	    gcif_uchar_t;
typedef signed char	    gcif_schar_t;
typedef unsigned short	    gcif_ushort_t;
typedef signed short	    gcif_sshort_t;

#define GCIF_ALIGN( v, a )  (( (((unsigned) (v))-1) | ((a)-1) )+1)

/* ======== GCIF_ARG ================================================== */
/* PURPOSE:
 *	One argument of the generalized call interface.  Each argument
 *  being passed will need one of these to describe it.  double's are
 *  always 8 byte aligned. Structures are aligned with either 4 or
 *  if they contain doubles with 8.  Everything else is 4.
 *  The size of any arg need not be of any multiple as it will be
 *  corrected when args are stuffed.
 */

typedef struct  GCIF_ARG {

    GCIF_ETYPE	gcif_type : 8;	/* arg type				*/
    unsigned	gcif_align : 8;	/* Alignment - 4, 8 or 16		*/
    unsigned	gcif_size : 16;	/* Call size				*/

} GCIF_ARG;


/* ======== GCIF_CALLDESC ============================================= */
/* PURPOSE:
 *	Call descriptor
 */

typedef struct  GCIF_CALLDESC {

    GCIF_ARG	      * gcif_args;	/* args				*/
    unsigned		gcif_nargs;	/* number of args		*/

    GCIF_ETYPE	        gcif_rettype;	/* return type			*/
    unsigned		gcif_retsize;	/* return size (for structs)	*/

    /* Filled in by gcif_filldesc if called.				*/
    unsigned		gcif_bytes;	/* Bytes required on call stack	*/
    unsigned		gcif_flags;	/* gencall flags		*/

    /* Filled in by user just before being called			*/
    void	     ** gcif_argv;	/* arg value pointers		*/
    int			gcif_nargv;	/* number of value pointers	*/

    void	      * gcif_retv;	/* Return value			*/

} GCIF_CALLDESC;


/* ======== gcif_Prefunc ============================================== */
/* PURPOSE:
 *	To be used with CallFromGeneralArgs as the prep func.  It takes
 *  the description of the arguments in GCIF_CALLDESC and generates
 *  a call frame for the function call described therin.
 *
 * RETURNS:
 * 	Nothing
 */

void	gcif_Prefunc(
    char	      * call_args,  /* Address of first arg to func	*/
    GCIF_CALLDESC     * args,	    /* As passed in CallFromGeneralArgs	*/
    int		        bytes,	    /* As passed in CallFromGeneralArgs	*/
    unsigned	      * p_flags     /* Pointer to the value "flags"	*/
);

/* ======== gcif_call ================================================= */
/* PURPOSE:
 *	Call a function with arguments described in the call description
 *  structure GCIF_CALLDESC. It also takes care of making a return
 *  address for functions returning structures.
 *
 * RETURNS:
 * 	Whatever func returns if it is not returning a struct.
 */

int gcif_call(
    GCIF_FUNC_T		  func,
    GCIF_CALLDESC	* args
);

void	gcif_filldesc(
    GCIF_CALLDESC	* args
);

#endif
